/* 
 * An implementation of printf that
 * - allows printing of 16-bit unicode characters and strings
 * - translates output to a specified character set
 *
 * "char8" is 8 bits and contains ISO-Latin-1 (or ASCII) values
 * "char16" is 16 bits and contains UTF-16 values
 * "Char" is char8 or char16 depending on whether CHAR_SIZE is 8 or 16
 *
 * Author: Richard Tobin
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#ifdef FOR_NSL
#include "lt-defs.h"
#endif

#include "charset.h"
#include "string16.h"
#include "stdio16.h"

/* When we return -1 for non-io errors, we set errno to 0 to avoid confusion */
extern int errno;

#define BufferSize 4096

typedef int WriteProc(void **handle, const unsigned char *buf, int count);

static int Vxprintf(WriteProc *write, void **handle, CharacterEncoding enc, 
		    const char *format, va_list args);
static int ConvertLatin1(const char8 *buf, int count, 
		       WriteProc *write, void **handle, CharacterEncoding enc);
static int ConvertUTF16(const char16 *buf, int count, 
		       WriteProc *write, void **handle, CharacterEncoding enc);
static int FileWriteBlock(void **handle, const unsigned char *buf, int count);
static int StringWriteBlock(void **handle, const unsigned char *buf,int count);

/* Output an ISO-Latin-1 buffer in the specified encoding */

static int ConvertLatin1(const char8 *buf, int count, 
		       WriteProc *write, void **handle, CharacterEncoding enc)
{
    unsigned char outbuf[BufferSize*2];
    int i, j;

    switch(enc)
    {
    case CE_ISO_8859_1:
	return write(handle, (unsigned char *)buf, count);
    case CE_UTF_8:
	for(i=j=0; i<count; i++)
	{
	    unsigned char c = buf[i];
	    if(c < 128)
		outbuf[j++] = c;
	    else
	    {
		outbuf[j++] = 0xc0 + (c >> 6);
		outbuf[j++] = 0x80 + (c & 0x3f);
	    }
	}
	return write(handle, outbuf, j);
    case CE_UTF_16B:
    case CE_ISO_10646_UCS_2B:
	for(i=j=0; i<count; i++)
	{
	    outbuf[j++] = 0;
	    outbuf[j++] = buf[i];
	}
	return write(handle, outbuf, count*2);
    case CE_UTF_16L:
    case CE_ISO_10646_UCS_2L:
	for(i=j=0; i<count; i++)
	{
	    outbuf[j++] = buf[i];
	    outbuf[j++] = 0;
	}
	return write(handle, outbuf, count*2);
    default:
	fprintf(stderr, "Bad output character encoding %d (%s)\n",
		enc,
		enc < CE_enum_count ? CharacterEncodingName[enc] : "unknown");
	errno = 0;
	return -1;
    }
}

/* Output a UTF-16 buffer in the specified encoding */

static int ConvertUTF16(const char16 *buf, int count, 
		       WriteProc *write, void **handle, CharacterEncoding enc)
{
    /* size is max for UTF-8 with saved first half */
    unsigned char outbuf[BufferSize*3 + 1];
    int i, j;
    static char16 first_half;	/* Eeek! this makes it non-reentrant */
    char32 big;

    switch(enc)
    {
    case CE_ISO_8859_1:
	for(i=0; i<count; i++)
	{
	    if(buf[i] < 256)	/* XXX */
		outbuf[i] = (unsigned char)buf[i];
	    else
		outbuf[i] = '?';
	}
	return write(handle, outbuf, count);
    case CE_UTF_8:
	for(i=j=0; i<count; i++)
	{
	    if(buf[i] < 0x80)
		outbuf[j++] = (unsigned char)buf[i];
	    else if(buf[i] < 0x800)
	    {
		outbuf[j++] = 0xc0 + (buf[i] >> 6);
		outbuf[j++] = 0x80 + (buf[i] & 0x3f);
	    }
	    else if(buf[i] >= 0xd800 && buf[i] <= 0xdbff)
		first_half = buf[i];
	    else if(buf[i] >= 0xdc00 && buf[i] <= 0xdfff)
	    {
		big = 0x10000 + 
		    ((first_half - 0xd800) << 10) + (buf[i] - 0xdc00);
		outbuf[j++] = 0xf0 + (big >> 18);
		outbuf[j++] = 0x80 + ((big >> 12) & 0x3f);
		outbuf[j++] = 0x80 + ((big >> 6) & 0x3f);
		outbuf[j++] = 0x80 + (big & 0x3f);
	    }
	    else
	    {
		outbuf[j++] = 0xe0 + (buf[i] >> 12);
		outbuf[j++] = 0x80 + ((buf[i] >> 6) & 0x3f);
		outbuf[j++] = 0x80 + (buf[i] & 0x3f);
	    }
	}
	return write(handle, outbuf, j);
    case CE_UTF_16B:
    case CE_ISO_10646_UCS_2B:
 	for(i=j=0; i<count; i++)
	{
	    outbuf[j++] = (buf[i] >> 8);
	    outbuf[j++] = (buf[i] & 0xff);

	}
	return write(handle, outbuf, count*2);
    case CE_UTF_16L:
    case CE_ISO_10646_UCS_2L:
 	for(i=j=0; i<count; i++)
	{
	    outbuf[j++] = (buf[i] & 0xff);
	    outbuf[j++] = (buf[i] >> 8);

	}
	return write(handle, outbuf, count*2);
    default:
	fprintf(stderr, "Bad output character encoding %d (%s)\n",
		enc,
		enc < CE_enum_count ? CharacterEncodingName[enc] : "unknown");
	errno = 0;
	return -1;
    }
}

static int FileWriteBlock(void **handle, const unsigned char *buf, int count)
{
    FILE *f = *handle;
    if(count == 0)
	return 0;
    return fwrite(buf, 1, count, f) == 0 ? -1 : 0;
}

static int StringWriteBlock(void **handle, const unsigned char *buf, int count)
{
    char *p = *handle;
    memcpy(p, buf, count);
    *handle = p + count;
    return 0;
}

int Fprintf(FILE *file, CharacterEncoding enc, const char *format, ...)
{
    va_list args;
    va_start(args, format);
    return Vfprintf(file, enc, format, args);
}

int Printf(CharacterEncoding enc, const char *format, ...)
{
    va_list args;
    va_start(args, format);
    return Vfprintf(stdout, enc, format, args);
}

int Sprintf(void *buf, CharacterEncoding enc, const char *format, ...)
{
    va_list args;
    va_start(args, format);
    return Vsprintf(buf, enc, format, args);
}

int Vfprintf(FILE *file, CharacterEncoding enc, const char *format, 
	     va_list args)
{
    void *handle = file;
    return Vxprintf(FileWriteBlock, &handle, enc, format, args);
}

int Vprintf(CharacterEncoding enc, const char *format, va_list args)
{
    void *handle = stdout;
    return Vxprintf(FileWriteBlock, &handle, enc, format, args);
}

int Vsprintf(void *buf, CharacterEncoding enc, const char *format, 
	     va_list args)
{
    int nchars;
    static char8 null = 0;

    nchars = Vxprintf(StringWriteBlock, &buf, enc, format, args);
    ConvertLatin1(&null, 1, StringWriteBlock, &buf, enc); /* null terminate */

    return nchars;
}

#define put(x) {nchars++; if(count == sizeof(buf)) {if(ConvertLatin1(buf, count, write, handle, enc) == -1) return -1; count = 0;} buf[count++] = x;}

int Vxprintf(WriteProc *write, void **handle, CharacterEncoding enc, 
	     const char8 *format, va_list args)
{
    char8 buf[BufferSize];
    int count = 0;
    int c, i, n, width, prec;
    char fmt[200];
    char8 val[200];
    const char8 *start;
    const char8 *p;
    const char16 *q;
    int mflag, pflag, sflag, hflag, zflag;
    int l, h, L;
    int nchars = 0;

    while((c = *format++))
    {
	if(c != '%')
	{
	    put(c);
	    continue;
	}

	start = format-1;
	width = 0;
	prec = -1;
	mflag=0, pflag=0, sflag=0, hflag=0, zflag=0;
	l=0, h=0, L=0;

	while(1)
	{
	    switch(c = *format++)
	    {
	    case '-':
		mflag = 1;
		break;
	    case '+':
		pflag = 1;
		break;
	    case ' ':
		sflag = 1;
		break;
	    case '#':
		hflag = 1;
		break;
	    case '0':
		zflag = 1;
		break;
	    default:
		goto flags_done;
	    }
	}
    flags_done:

	if(c == '*')
	{
	    width = va_arg(args, int);
	    c = *format++;
	}
	else if(c >= '0' && c <= '9')
	{
	    width = c - '0';
	    while((c = *format++) >= '0' && c <= '9')
		width = width * 10 + c - '0';
	}

	if(c == '.')
	{
	    c = *format++;
	    if(c == '*')
	    {
		prec = va_arg(args, int);
		c = *format++;
	    }
	    else if(c >= '0' && c <= '9')
	    {
		prec = c - '0';
		while((c = *format++) >= '0' && c <= '9')
		    prec = prec * 10 + c - '0';
	    }
	    else
		prec = 0;
	}

	switch(c)
	{
	case 'l':
	    l = 1;
	    c = *format++;
	    break;
	case 'h':
	    h = 1;
	    c = *format++;
	    break;
	case 'L':
	    L = 1;
	    c = *format++;
	    break;
	}

	if(format - start + 1 > sizeof(fmt))
	{
	    fprintf(stderr, "Printf: format specifier too long");
	    errno = 0;
	    return -1;
	}

	strncpy(fmt, start, format - start);
	fmt[format - start] = '\0';

	/* XXX should check it fits in val */

	switch(c)
	{
	case 'n':
	    *va_arg(args, int *) = nchars;
	    break;
	case 'd':
	case 'i':
	case 'o':
	case 'u':
	case 'x':
	case 'X':
	    if(h)
		sprintf(val, fmt, va_arg(args, int)); /* promoted to int */
	    else if(l)
		sprintf(val, fmt, va_arg(args, long));
	    else
		sprintf(val, fmt, va_arg(args, int));
	    for(p=val; *p; p++)
		put(*p);
	    break;
	case 'f':
	case 'e':
	case 'E':
	case 'g':
	case 'G':
	    if(L)
		sprintf(val, fmt, va_arg(args, long double));
	    else
		sprintf(val, fmt, va_arg(args, double));
	    for(p=val; *p; p++)
		put(*p);
	    break;
	case 'c':
	    put(va_arg(args, int));
	    break;
	case 'p':
	    sprintf(val, fmt, va_arg(args, void *));
	    for(p=val; *p; p++)
		put(*p);
	    break;
	case '%':
	    put('%');
	    break;
	case 's':
	    if(l)
	    {
		static char16 sNULL[] = {'(','N','U','L','L',')',0};
#if CHAR_SIZE == 16
	    string:
#endif
		q = va_arg(args, char16 *);
		if(!q)
		    q = sNULL;
		n = strlen16(q);
		if(prec >= 0 && n > prec)
		    n = prec;
		if(n < width && !mflag)
		    for(i=width-n; i>0; i--)
			put(' ');
		if(ConvertLatin1(buf, count, write, handle, enc) == -1)
		    return -1;
		count = 0;
		if(ConvertUTF16(q, n, write, handle, enc) == -1)
		    return -1;
		nchars += n;
	    }
	    else
	    {
#if CHAR_SIZE == 8
	    string:
#endif
		p = va_arg(args, char8 *);
		if(!p)
		    p = "(null)";
		n = strlen(p);
		if(prec >= 0 && n > prec)
		    n = prec;
		if(n < width && !mflag)
		    for(i=width-n; i>0; i--)
			put(' ');
		for(i=0; i<n; i++)
		    put(p[i]);
	    }
	    if(n < width && mflag)
		for(i=width-n; i>0; i--)
		    put(' ');
	    break;
	case 'S':
	    goto string;
	default:
	    fprintf(stderr, "unknown format character %c\n", c);
	    errno = 0;
	    return -1;
	}
    }

    if(count > 0)
	if(ConvertLatin1(buf, count, write, handle, enc) == -1)
	    return -1;

    return nchars;
}

#ifdef test

int main(int argc, char **argv)
{
    short s=3;
    int n, c;
    char16 S[] = {'w', 'o', 'r', 'l', 'd', ' ', '' & 0xff, 0xd841, 0xdc42, 0};

    n=Printf(CE_ISO_8859_1, argv[1], s, 98765432, &c, 5.3, 3.2L, "hello", S);
    printf("\nreturned %d, c=%d\n", n, c);
    n=Printf(CE_UTF_8, argv[1], s, 98765432, &c, 5.3, 3.2L, "hello", S);
    printf("\nreturned %d, c=%d\n", n, c);
    n=Printf(CE_UTF_16L, argv[1], s, 98765432, &c, 5.3, 3.2L, "hello", S);
    printf("\nreturned %d, c=%d\n", n, c);
    n=Printf(CE_UTF_16B, argv[1], s, 98765432, &c, 5.3, 3.2L, "hello", S);
    printf("\nreturned %d, c=%d\n", n, c);

    return 0;
}

#endif

